import pandas as pd
import numpy as np
import psycopg2 # needed to connect to redshift
import sqlalchemy
from sqlalchemy import event, create_engine
from sqlalchemy import types

def create_table_in_redshift(table_name, 
                             columns, 
                             credentials,
                             primary_key = None, 
                             sql_type='postgresql',
                             schema='schema_name', 
                             delete_existing=False):
    
    """
    Create new table in redshift    
    Prereq: psycopg2 package
    Inputs
        table_name (str) - name of table
        columns (list) - [('id', 'BIGINT'), ('col2', 'DATE')]
        credentials (dict) - Redshift credentials dict(dbname, host, port, user, password)
        primary_key (str) - default=None, column name of primary key e.g. 'id'
        sql_type='postgresql'  or 'mysql'
        schema (str) - default = schema_name. schema where we save the table
        delete_existing (bool) - default=False. WARNING: will delete existing table and replace
    Outputs
        new table created in Redshift
    """
    
    # convert column list to SQL language
    cols_str = ',\n'.join([' '.join(i) for i in columns])

    # delete existin table statement
    if delete_existing == True:
        delete = f"DROP TABLE IF EXISTS {schema}.{table_name};"
    else:
        delete = ""

    # add primary key row or not
    if primary_key != None:
        primary_key_statement = f', PRIMARY KEY({primary_key})'
    else:
        primary_key_statement = ''

    # SQL query
    sql = f"""
    {delete}
    CREATE TABLE {schema}.{table_name} (
    {cols_str}
    {primary_key_statement}
    );
    """
    # print(sql)
   

    if sql_type == 'postgresql':
        # connect to redshift database
        conn=psycopg2.connect(**credentials)    
        
    elif sql_type == 'mysql':
        conn=pymysql.connect(**credentials)                    

    # # connect to redshift database
    # conn=psycopg2.connect(**credentials)    
    # create a new cursor
    cur = conn.cursor()
    # execute the SQL statement
    cur.execute(sql)
    # commit the changes to the database
    conn.commit()
    # close communication with the database
    cur.close()    
    # close the connection
    conn.close()
    
    return f'created table {schema}.{table_name}'



def insert_pandas_data_to_redshift_table(df, 
                                         table_name, 
                                         credentials, 
                                         chunksize=500,
                                         schema="schema_name"):
    """
    Insert pandas data to existing redshift table
    prereq: conda install sqlalchemy
    Inputs
        df - pandas dataframe
        table_name (str)
        credentials (dict) - redshift credentials
        chunksize(int) - default = 500, how many rows to upload at a time
        schema (str) - default = "schema_name"
    Output
        insert data into table in redshift
    """    
    user = credentials['user']
    host = credentials['host']
    dbname = credentials['dbname']
    port = credentials['port']
    password = credentials['password']   

    engine = sqlalchemy.create_engine(f'postgresql://{user}:{password}@{host}:{port}/{dbname}')

    # this supposedly makes uploading faster
    dtyp = {c:types.VARCHAR(df[c].str.len().max())
            for c in df.columns[df.dtypes == 'object'].tolist()}

    with engine.connect() as conn:
        df.to_sql(table_name, con=engine, schema=schema, if_exists='append', 
                         index=False, index_label=None, chunksize=chunksize, 
                          dtype=dtyp, method='multi')    
        print(f"saved to {schema}.{table_name}") # <- just to keep track of the process

