from collections import defaultdict
import pandas as pd
import numpy as np
import psycopg2
import pymysql
from tqdm import tqdm
from itertools import chain
from decouple import Config, RepositoryEnv
from sql_pylib.load_from_redshift_to_pandas import load_from_redshift_to_pandas
from sql_pylib import database_download_upload as db

# +
class SQLBuilder:
    def __init__(self, 
                 credentials=None, 
                 sql_type='postgresql'):
        """
        credentials - dictionary of redshift creds or path to .env file
        sql_type='postgresql' or 'mysql'
        """
                
        # if path to the .env file
        if type(credentials) == str:
            print(f"loading credentials from {credentials}")
            # ENV PATH
            env_config = Config(RepositoryEnv(credentials))
            credentials = dict(
                dbname=env_config.get('NAME_DB'),
                host=env_config.get('HOST_DB'),
                port=env_config.get('PORT_DB'),
                user=env_config.get('USER_DB'),
                password=env_config.get('PASSWORD_DB')    
                )
        
        # Initial values for building SQL queries
        self.query = "SELECT"        
        self.aliases = [] # defaultdict() #{source_query: alias}
        self.alias_num = 1
        self.columns = []
        self.has_cols = False # whether or not we have columns added already
        self.credentials = credentials
        self.sql_type = sql_type               
        
        return None
    
    ########################################
    # AUTOMATED SQL QUERY GENERATION
    ########################################
    
    def select(self, columns, distinct=True):
        """        
        columns (list or str) - list of columns, e.g. ["col1", "col2"] or
                            actual query e.g. "col1, col2, col1+col2 AS col3"
        distinct (bool) - default=True. whether to select distinct columns or not
        
        Can do multile select e.g. sql.select(cols1).select(cols2)
        """
        # whether the select has cols or not
        has_cols = self.has_cols
        
        # if query does not yet have SELECT at the start        
        if not has_cols:
            if distinct:
                self.query += " DISTINCT"            
        
        if type(columns) == str:
            columns = columns.split(",")
            
        # keep appending columns
        self.columns += columns 
            
        added_comma_for_new_cols = False
        for col in columns:
            # if query has cols already, append a "," at the start of the columns
            if has_cols and not added_comma_for_new_cols:                            
                self.query += ",\n " + str(col).strip() + ","
                added_comma_for_new_cols = True
            else:                            
                self.query += "\n " + str(col).strip() + ","                
    
        self.query = self.query[:-1] # remove last ","        
        
        # first select has cols now
        self.has_cols = True
        
        return self
    

    def from_table(self, source, alias=None):
        """
        source (str/SQLBuilder) - table name, e.g. "schema_name.table_name" or it can be
                    a query, e.g. "SELECT * FROM table" or an SQLBuilder query object
        alias (str) - default=None. alias of this table, otherwise, automatically add new one
        """
        
        # if source is SQLBuilder object
        if type(source) == type(self):
            source = source.query # get the string query version
        
        if alias == None:
            alias = f"table_{self.alias_num}"
            self.alias_num += 1 # for new sources
            
        # add parentheses if SELECT in table
        if "select" in source.lower():
            source = f"(\n{source}\n)"
            
        # append alias for this source                  
        self.aliases += [alias]                    
        
        # append alias to source
        source = f"{source} {alias}"
        
        self.query += f"\nFROM\n{source}"
        return self

    def where(self, conditions):
        """
        conditions (str) - 'col1 > col2 AND col1 = 0'
        """
        
        self.query += f"\nWHERE\n  ({conditions})"
        return self
    
    def having(self, conditions):
        """
        conditions (str) - 'col1 > col2 AND col1 = 0'
        """
        
        self.query += f"\nHAVING\n  ({conditions})"
        return self    

    def condition_or(self, conditions):
        """
        Add conditions joined by OR
        conditions (str) - 'col1 > col2'
        """
        
        self.query += f"\nOR ({conditions})"
        return self
    
    def condition_and(self, conditions):
        """
        Add conditions joined by AND
        conditions (str) - 'col1 > col2'
        """
        
        self.query += f"\nAND ({conditions})"
        return self    
        
    
    def join(self, source, on=None, 
             left_on=None, right_on=None,             
             on_left=None, on_right=None, 
             how="left", alias=None):
        """
        source (str/SQLBuilder) - table or SQL query
        on (list) - column(s) to join if same column names for left and right tables
        
        left_on (str/list) - column(s) to join on left table, e.g. ["col1", "col2"]
        right_on (str/list) - column(s) to join on right table, e.g. ["col1", "col2"]        
        
        on_left (str/list) - alternative arg. column(s) to join on left table, e.g. ["col1", "col2"]
        on_right (str/list) - alternative arg. column(s) to join on right table, e.g. ["col1", "col2"]
        
        how - "left", "right", or "inner"
        alias (str) - default = None. alias of the new table to join
        """        
        
        # for strings
        if type(on) == str:
            on = [on]
        
        if type(left_on) == str:
            left_on = [left_on]
            
        if type(right_on) == str:
            right_on = [right_on] 
            
        if type(on_left) == str:
            on_left = [on_left]        
            
        if type(on_right) == str:
            on_right = [on_right]                    
        
        # convert SQLBuilder source to query string
        if type(source) == type(self):
            source = source.query # get the string query version        
        
        # get alias of existing table / source
        alias_left = self.aliases[-1]
        self.alias_num += 1 # for new sources                   
        
        # add parentheses if SELECT in table
        if "select" in source.lower():
            source = f"(\n{source}\n)"        
        
        # add alias to new table                
        if alias is None:
            alias_right = f"table_{self.alias_num}"
            self.alias_num += 1 # for new sources      
        else:
            alias_right = alias
        source += f" {alias_right}"          
        
        how = how.upper()
        self.query += f"\n{how} JOIN {source}\n  ON "
        
        # if on - same column names
        if on is not None:
            for i in range(len(on)):
                left_right = on[i]
                self.query += f"{alias_left}.{left_right} = {alias_right}.{left_right}"
                self.query += "\n"           

                if i+1 < len(on):
                    self.query += "  AND "                
        
        elif left_on is not None and right_on is not None:        
            for i in range(len(left_on)):
                left = left_on[i]
                right = right_on[i]
                self.query += f"{alias_left}.{left} = {alias_right}.{right}"
                self.query += "\n"           

                if i+1 < len(left_on):
                    self.query += "  AND "
                    
        elif on_left is not None and on_right is not None:        
            for i in range(len(on_left)):
                left = on_left[i]
                right = on_right[i]
                self.query += f"{alias_left}.{left} = {alias_right}.{right}"
                self.query += "\n"           

                if i+1 < len(on_left):
                    self.query += "  AND "                    
                    
        else:
             assert on is not None or (left_on is not None and right_on is not None), f"Missing argument 'on' or ('left_on' and 'right_on')"
                
        return self

    def groupby(self, cols):
        """
        cols (list) - columns to group by e.g. ["col1", "col2"] or string "col1"
        """
        
        self.query += "\n"
        self.query += "GROUP BY"        

        if type(cols) == str:
            cols = [cols]
        
        for i in range(len(cols)):
            col = cols[i]
            self.query += f"\n  "
            self.query += f"{col}"
            
            if i+1 < len(cols):
                self.query += ", "                        
        return self
    
    
    def orderby(self, cols, how=["asc"]):
        """
        cols (list) - columns to sort
        how (list) - default = ["asc"], list of "asc" and "desc" corresponding to each column    
        """
        
        order = "\nORDER BY\n"
        for i in range(len(cols)):
            col_ = cols[i]
            how_ = how[i].upper()
            
            order += f"  {col_} {how_}"
            if i+1 < len(cols):
                order += ", "
                
        self.query += order
        return self
    
    
    def union(self, table):
        """
        table (str) - query or table to union on
        """
        
        # if SQLBuilder object, get the query
        if type(table) == type(self):
            table = table.query # get the string query version        
        
        # add parentheses if SELECT in table
        if "select" in table.lower():
            table = f"(\n{table}\n)"        
        
        self.query += f"\n\nUNION ALL\n\n{table}"
        return self
    
    
    def limit(self, num_rows):
        """
        num_rows (int) - how many rows to show
        """
        self.query += f"\nLIMIT {num_rows}"
        return self    
    
    ######################################## 
    # CHECK COLUMN NAMES AND TABLE NAMES
    ########################################    
    
    def get_table_names(self, schema):
        """
        Get table names given schema
        Input
            schema (str) - schema name
        Returns
            query for list of table names in schema
        """
        query = f"""
        select t.table_name
        from information_schema.tables t
        where t.table_schema = '{schema}'
              and t.table_type = 'BASE TABLE'
        order by t.table_name;
        """

        return query
    
    def get_column_names(self, table_name, schema=None):
        """
        Get column names given table name list
        Input
            table_name (str) - single table name
            schema (str) - optional. schema to use
            
        Returns
            query for columns and descriptions
        """
        
        sql_type = self.sql_type
        
        if sql_type == 'postgresql':
            if type(schema) == str:
                schema_query = f"AND table_schema = '{schema}'"
            else:
                schema_query = ""

            query = f"""
            SELECT
                table_schema
                , table_name
                , ordinal_position
                , column_name
                , column_default
                , data_type
                , is_nullable    
            FROM 
               information_schema.columns
            WHERE 
                table_name = '{table_name}'
                {schema_query}
            ORDER BY ordinal_position      
            """
            
        elif sql_type == 'mysql':
            query = f"""
            SELECT COLUMN_NAME AS column_name
            FROM INFORMATION_SCHEMA.COLUMNS 
            WHERE 
                TABLE_SCHEMA = Database()
            AND TABLE_NAME = '{table_name}'            
            """
        
        return query            
        
        
        
    ######################################## 
    # RUNNING AND DOWNLOADING QUERIES
    ########################################   
      
    # RUN any query
    def run_query(self, query):
        """
        Run query without downloading
        sql_type='postgresql' or 'mysql
        """
        credentials = self.credentials
        sql_type = self.sql_type
        
        if sql_type == 'postgresql':
            # connect to redshift database
            conn=psycopg2.connect(**credentials)    

        elif sql_type == 'mysql':
            conn=pymysql.connect(**credentials)    
            
        # # connect to redshift database
        # conn=psycopg2.connect(**credentials)    
        # create a new cursor
        cur = conn.cursor()
        # execute the SQL statement"
        cur.execute(query)
        # commit the changes to the database
        conn.commit()
        # close communication with the database
        cur.close()    
        # close the connection
        conn.close()
        # print('query executed')            
        
        
    # Download any query
    def download(self, query):
        """
        Download queried table        
        query (str/SQLBuilder) - query string or SQLBuilder object
        """
        
        # whether mysql or postgresql
        sql_type = self.sql_type
        
        # if SQLBuilder then get the text form
        if type(query) == type(self):
            query = query.query
        
        df = load_from_redshift_to_pandas(query, self.credentials, sql_type)             
        return df    
    
    # same function as download
    # Download any query    
    def download_query(self, query):
        """
        Download queried table        
        query (str/SQLBuilder) - query string or SQLBuilder object
        """
        df = self.download(query)
        return df
    
    
    
    ######################################## 
    # CREATING OR INSERTING TO TABLE
    ########################################    
    
    # create new empty table
    def create_table(self,
                     table_name,
                     columns, 
                     primary_key = None, 
                     sql_type='postgresql',
                     schema='schema_name', 
                     delete_existing=False):
        
        """
        Create new table in redshift    
        Inputs
            table_name (str) - name of table
            columns (list) - [('id', 'BIGINT'), ('col2', 'DATE')]
            primary_key (str) - default=None, column name of primary key e.g. 'id'
            sql_type='postgresql'  or 'mysql'
            schema (str) - default = schema_name. schema where we save the table
            delete_existing (bool) - default=False. WARNING: will delete existing table and replace
        Outputs
            new table created in Redshift
        """        
    
        db.create_table_in_redshift(table_name, 
                                    columns, 
                                    self.credentials,
                                    primary_key = primary_key,
                                    delete_existing = delete_existing,
                                    schema=schema
                                   )    
    
    
    # create and upload new table from pandas dataframe
    def upload_table(self, 
                     df, 
                     table_name, 
                     schema = 'schema_name', 
                     chunksize = 1000,                     
                     delete_existing = False,
                     varchar_len = 1024,
                     primary_key = None,                     
                    ):
        """
        creates new table on redshift based on pandas dataframe
        upload data to table on redshift
        
        Args:
            df - Pandas dataframe
            table_name - string. name of table on redshift
            schema - string. default = 'schema_name'. schema where to create table.
            chunk_size - int. default = 1000. how many rows to upload at a time
            delete_existing - default = False. whether to delete the existing table
            varchar_len - default = 1024. the lenght of VARCHAR(<length>)
            primary_key - default = None. string. which column to use as primary key
        Returns
            None
        """
        
        # mapping for converting pandas datatypes to redshift datatypes
        data_types_map = {
            np.dtype('O'):f"VARCHAR({varchar_len})",
            np.dtype('int64'):"BIGINT",
            np.dtype('int32'):"BIGINT",
            np.dtype('float32'):"FLOAT",
            np.dtype('float64'):"FLOAT",    
            np.dtype('<M8[ns]'): "DATETIME",
        }
        
        # conver cols to to "<col>"
        cols = [f"""\"{c}\"""" for c in df.columns]

        # convert pandas datatypes to redshift datatypes
        dtypes_sql = [data_types_map[i] for i in df.dtypes.values]    

        # generate tuple of column names and dtypes
        # columns = list(zip(*[list(df.dtypes.keys()), dtypes_sql]))
        columns = list(zip(*[cols, dtypes_sql]))
        
        # create new empty table based on generted columns from the pd dataframe
        db.create_table_in_redshift(table_name, 
                                    columns, 
                                    self.credentials,
                                    primary_key = primary_key,
                                    delete_existing = delete_existing,
                                    schema=schema
                                   )    
        
        # insert to created table
        db.insert_pandas_data_to_redshift_table(df, 
                                                 table_name, 
                                                 self.credentials, 
                                                 chunksize=chunksize,
                                                 schema=schema
                                               )
        
    # insert to existing table
    def insert_to_table(self,
                        df, 
                        table_name,
                        schema="schema_name",                       
                        chunksize=1000, ):  
        """
        Insert pandas data to existing redshift table
        Inputs
            df - pandas dataframe
            table_name (str)
            chunksize(int) - default = 1000, how many rows to upload at a time
            schema (str) - default = "schema_name"
        Output
            insert data into table in redshift
        """            
        # insert to created table
        db.insert_pandas_data_to_redshift_table(df, 
                                                 table_name, 
                                                 self.credentials, 
                                                 chunksize=chunksize,
                                                 schema=schema
                                               )
    
    ######################################## 
    # GENERATING QUERY FROM EXISTING TABLE
    ########################################        
    
    # GENERAL QUERY
    def make_query(self,
                    schema='schema_name', 
                    table_name=None,
                    cols = None,
                    filters=None,
                    limit=None):
        """
        Load general query
        
        Args:
            schema - schema name
            cols - None, if None, download all accessible cols. Otherwise, provide a list, ['col1', 'col2']
            filters = None, string, apply certain filters
            limit - None, can change to how many rows to query, e.g. 5
            
        Returns:
            query object. To get text query, run <output_var>.query
        """

        # # get cols query
        q = SQLBuilder(sql_type=self.sql_type).get_column_names(table_name, schema=schema)

        if cols is None:
            # cols = ["*"] # everything - will throw an error if no permissions on some cols
            cols = list(load_from_redshift_to_pandas(q, self.credentials, self.sql_type).column_name.values.flatten())
            
        table = f"{schema}.{table_name}"
        
        if len(cols) == 0:
            print(f"Can't find table {table}")
            return None
        
        sql = (SQLBuilder().select(cols)
                      .from_table(table)
                 )

        if filters is not None:
            sql = (sql.where(filters))

        if limit is not None:
            sql = sql.limit(limit)
        
        return sql
        
    
    ######################################## 
    # AUTOMATED UNION
    ########################################    
    
    #### FOR AUTOMATED UNION
    def get_table_columns(self, table):
        """
        Args
            table - either table name or SQLBuilder object
        Returns
            list of columns for that table        
        """
        t_ = SQLBuilder().select("*").from_table(table)
        df_ = self.download_query(t_.limit(1))
        return list(df_.columns)


    def make_union_query(self, input_tables:list, cols_list = "infer", conv_to_text = True):
        """
        Args
            input_tables (list) - list of input tables either table names or SQLBuilder objects
            cols_list (list) - default='infer' which infers the columns, otherwise provide a list of columns [['c1', 'c2'], ['c1', 'c3']]
            conv_to_text (bool) - default = True which converts all columns to texts if there are tables with different datatypes for the same column
        Returns
            query to union the different tables
        Notes
            converts all columns to text
            need to post process after to convert back to desired datatypes
        """

        # get columns per table
        if cols_list == 'infer':
            cols_list = []
            for i in tqdm(range(len(input_tables))):
                t = input_tables[i]
                
                # if table has already been created
                # this is faster
                if type(t) == str:
                    c = list(self.download_query(self.get_column_names(t.split('.')[1]))['column_name'].values.flatten())
                    c = sorted(c)
                
                # if the table is a query SQLBuilder
                else:
                    c = self.get_table_columns(t)
                    c = sorted(c)
                    
                cols_list.append(c) 

        # find the union of columns and sort in alphabetical order
        all_cols = sorted(list(set(chain(*cols_list))))    

        # generate a SELECT statement with NULLS if col does not exist in table for each table
        # sorted in alphabetical order
        # for loop union with corresponding select statements

        for i in range(len(input_tables)):

            # get specific table from tables list
            t = input_tables[i]

            # which cols need to be appended - dont exist in the table
            null_cols = list(set(all_cols) - set(cols_list[i]))

            # append NULL if part of null_cols

            if conv_to_text:
                # make everything text
                cols_ = [f"NULL AS {c}" if c in null_cols else f"{c}::TEXT AS {c}" for c in all_cols]
            else:
                cols_ = [f"NULL AS {c}" if c in null_cols else c for c in all_cols]            

            # if first table
            if i == 0:
                q_union = (SQLBuilder()
                           .select(cols_)
                           .from_table(t)
                          ) 
            # for succeeding tables
            elif i > 0:
                q_ = (SQLBuilder()
                       .select(cols_)
                       .from_table(t)
                      )
                q_union = q_union.union(q_.query)

        return q_union  
    
    ########################################    
    # MONITORING RUNNING QUERIES
    ########################################
    
    def check_running_processes(self):
        query = """select pid, starttime, duration,
                    trim(user_name) as user,
                    trim (query) as querytxt
                    from stv_recents
                    where status = 'Running';"""
        return query
    
    # TERMINATING RUNNING QUERIES
    def terminate_process(self, process_id):
        q = f"SELECT pg_terminate_backend({process_id})"
        return q    
        
        
        